#!/usr/bin/env python3

# TEST:source "$^CURRENT_DIRNAME/../lib/FC_Solve/__init__.py"
from FC_Solve import FC_Solve_Suite
import re
import os
import unittest


class MyTests(unittest.TestCase):
    def _eq(self, x, y, blurb):
        return self.assertEqual(x, y, blurb)

    def test_resume_solution(self):
        testname = "With RunIndef - "

        fcs = FC_Solve_Suite(self)

        limit = 10
        fcs.limit_iterations(limit)

        ret = fcs.solve_board(
            """4C 2C 9C 8C QS 4S 2H
5H QH 3C AC 3H 4H QD
QC 9S 6H 9H 3S KS 3D
5D 2S JC 5C JH 6D AS
2D KD TH TC TD 8D
7H JS KH TS KC 7C
AH 5S 6S AD 8H JD
7S 6C 7D 4D 8S 9D
"""
        )

        iters_count_ok = 1

        while (ret == 5):
            if (fcs.get_num_times() != limit):
                iters_count_ok = 0

            limit += 10
            fcs.limit_iterations(limit)
            ret = fcs.resume_solution()

        # TEST
        self._eq(ret, 0, testname + "State was successfully solved.")

        # TEST
        self._eq(iters_count_ok, 1,
                 testname + "Iters count was OK throughout the solution.")

    def test_num_states_in_collection_after_recycle(self):
        testname = "NumStatesInCol-After-Recycle - "

        fcs = FC_Solve_Suite(self)

        # MS-Freeceel board No. 24.
        fcs.solve_board(
            """4C 2C 9C 8C QS 4S 2H
5H QH 3C AC 3H 4H QD
QC 9S 6H 9H 3S KS 3D
5D 2S JC 5C JH 6D AS
2D KD TH TC TD 8D
7H JS KH TS KC 7C
AH 5S 6S AD 8H JD
7S 6C 7D 4D 8S 9D
"""
        )

        # TEST
        self._eq(fcs.get_num_times(), 137, testname + "Get num times is OK.")

        # TEST
        self._eq(fcs.get_num_states_in_col(), 191,
                 testname + "Num-states-in-collection is OK.")

        fcs.recycle()

        # TEST
        self._eq(fcs.get_num_times(), 0,
                 testname + "Get num times immediately after recycle.")

        # TEST
        self._eq(fcs.get_num_states_in_col(), 0, testname +
                 "Num-states-in-collection immediately after recycle.")

        # MS-Freeceel board No. 24.
        fcs.solve_board(
            """4C 2C 9C 8C QS 4S 2H
5H QH 3C AC 3H 4H QD
QC 9S 6H 9H 3S KS 3D
5D 2S JC 5C JH 6D AS
2D KD TH TC TD 8D
7H JS KH TS KC 7C
AH 5S 6S AD 8H JD
7S 6C 7D 4D 8S 9D
"""
        )

        # TEST
        self._eq(fcs.get_num_times(), 137, testname +
                 "Get num times after recycle.")

        # TEST
        self._eq(fcs.get_num_states_in_col(), 191,
                 testname + "Num-states-in-collection after recycle.")

    def test_num_states_in_collection_after_unsolved(self):
        testname = "NumStatesInCol-After-unsolved - "

        tags_str = os.getenv('FCS_TEST_TAGS')
        if not tags_str:
            tags_str = ''

        if re.search(r'\bfc_only\b', tags_str):
            return

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line("bakers_game", ['-g', 'bakers_game'])

        # MS-Freeceel board No. 10
        fcs.solve_board(
            """5S KD JC TS 9D KH 8D
5H 2S 9H 7H TD AD 6D
6H QD 6C TC AH 8S TH
6S 2D 7C QC QS 7D 3H
5D AS 7S KC 3D AC
4D 9C QH 4H 4C 5C
2H 3S 8H 9S JS 4S
JH JD 3C KS 2C 8C
"""
        )

        # TEST
        self._eq(fcs.get_num_times(), 3436, testname + "Get num times is OK.")

        # TEST
        self._eq(fcs.get_num_states_in_col(), 3436,
                 testname + "Num-states-in-collection is OK.")

    def test_resume_solution_with_flares(self):
        testname = "-l ve on iterative limiting - "

        fcs = FC_Solve_Suite(self)

        step = 1000
        hard_limit = 100000
        limit = step
        # TEST*$input_cmd_line
        fcs.input_cmd_line("video-editing", ['-l', 'video-editing'])
        fcs.limit_iterations(limit)

        # MS deal No. 124
        ret = fcs.solve_board(
            """7S AS 2C QS TH QD 7D
5D 8D 9D JH QH 5C JD
6D TD 5H 2S 6S TC KS
TS 4S 3D 9C 3C KD 7C
6H 5S 9H 6C KC AH
AC 4C 8S 2D QC JS
9S KH 8C 4D 7H 4H
2H 3S 8H AD 3H JC
"""
        )

        iters_count_ok = 1

        while (ret == 5 and limit < hard_limit):
            if (fcs.get_num_times() != limit):
                iters_count_ok = 0

            limit += step
            fcs.limit_iterations(limit)
            ret = fcs.resume_solution()

        # TEST
        self._eq(ret, 0, testname + "State was successfully solved.")

        # TEST
        self._eq(iters_count_ok, 1,
                 testname + "Iters count was OK throughout the solution.")


if __name__ == "__main__":
    # plan(14)
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)



#!/usr/bin/env python3
#
# find-freecell-deal-index.py - a program to find out the Freecell deal index
# based on the initial cards layout.
#
# Copyright by Shlomi Fish, 2018
#
# Licensed under the MIT/Expat License.

import sys
from make_pysol_freecell_board import find_index_main
from find_index_lib_py import ffi, lib


def _to_bytes(s):
    if sys.version_info > (3,):
        return bytes(s, 'UTF-8')
    else:
        return s


def find_ret(ints):
    obj = lib.fc_solve_user__find_deal__alloc()
    lib.fc_solve_user__find_deal__fill(
        obj, _to_bytes("".join(["%-10d" % x for x in ints])))
    ret = int(ffi.string(lib.fc_solve_user__find_deal__run(
        obj, b'1', _to_bytes("%d" % ((1 << 33) - 1)))))
    lib.fc_solve_user__find_deal__free(obj)
    return ret


if __name__ == "__main__":
    sys.exit(find_index_main(sys.argv, find_ret))



#!/usr/bin/env python3

import subprocess
import os
import unittest

py_prog = os.environ['FCS_SRC_PATH'] + '/board_gen/transpose-freecell-board.py'


def _normalize_lf(txt):
    return txt.replace("\r\n", "\n")


class MyTests(unittest.TestCase):
    def mytest(self, input_text, want_out, msg):
        global py_prog
        process = subprocess.Popen(
            ['python3', py_prog, '-'],
            shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        process.stdin.write(bytes(input_text, 'utf_8'))
        process.stdin.close()
        got_out = process.stdout.read().decode('utf-8')

        self.assertEqual(_normalize_lf(got_out), _normalize_lf(want_out), msg)

    def test_main(self):
        # plan(6)

        # TEST
        self.mytest(
            """Foundations:
Freecells:
4C 5H QC 5D 2D 7H AH 7S
2C QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD KC 8H 8S
4S 4H KS 6D 8D 7C JD 9D
2H QD 3D AS
""",
            """Foundations:
Freecells:
: 4C 2C 9C 8C QS 4S 2H
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS KC 7C
: AH 5S 6S AD 8H JD
: 7S 6C 7D 4D 8S 9D
""",
            "Freecell MS deal 24 initial",
        )

        # TEST
        self.mytest(
            """Foundations:
Freecells:
4C 5H QC 5D 2D 7H AH 7S
2H QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD KC 8H 8S
4S 4H KS 6D 8D 7C JD 9D
   QD 3D AS
      2C
""",
            """Foundations:
Freecells:
: 4C 2H 9C 8C QS 4S
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D 2C
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS KC 7C
: AH 5S 6S AD 8H JD
: 7S 6C 7D 4D 8S 9D
""",
            "With longer column."
        )

        # TEST
        self.mytest(
            """Foundations:
Freecells:
4C 5H QC 5D 2D 7H AH 7S
2H QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD    8H 8S
   4H KS 6D 8D    JD
   QD 3D AS       KC
      2C          9D
      4S
      7C
""",
            """Foundations:
Freecells:
: 4C 2H 9C 8C QS
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D 2C 4S 7C
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS
: AH 5S 6S AD 8H JD KC 9D
: 7S 6C 7D 4D 8S
""",
            "With two longer columns."
        )

        # TEST
        self.mytest(
            """Freecells:
4C 5H QC 5D 2D 7H AH 7S
2H QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD    8H 8S
   4H KS 6D 8D    JD
   QD 3D AS       KC
      2C          9D
      4S
      7C
""",
            """Freecells:
: 4C 2H 9C 8C QS
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D 2C 4S 7C
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS
: AH 5S 6S AD 8H JD KC 9D
: 7S 6C 7D 4D 8S
""",
            "Without a foundations line"
        )

        # TEST
        self.mytest(
            """Foundations:
4C 5H QC 5D 2D 7H AH 7S
2H QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD    8H 8S
   4H KS 6D 8D    JD
   QD 3D AS       KC
      2C          9D
      4S
      7C
""",
            """Foundations:
: 4C 2H 9C 8C QS
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D 2C 4S 7C
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS
: AH 5S 6S AD 8H JD KC 9D
: 7S 6C 7D 4D 8S
""",
            "Without a Freecells line"
        )

        # TEST
        self.mytest(
            """4C 5H QC 5D 2D 7H AH 7S
2H QH 9S 2S KD JS 5S 6C
9C 3C 6H JC TH KH 6S 7D
8C AC 9H 5C TC TS AD 4D
QS 3H 3S JH TD    8H 8S
   4H KS 6D 8D    JD
   QD 3D AS       KC
      2C          9D
      4S
      7C
""",
            """: 4C 2H 9C 8C QS
: 5H QH 3C AC 3H 4H QD
: QC 9S 6H 9H 3S KS 3D 2C 4S 7C
: 5D 2S JC 5C JH 6D AS
: 2D KD TH TC TD 8D
: 7H JS KH TS
: AH 5S 6S AD 8H JD KC 9D
: 7S 6C 7D 4D 8S
""",
            "With neither a Freecells line nor a Foundations line."
        )


if __name__ == '__main__':
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)



#!/usr/bin/env python3

# TEST:source "$^CURRENT_DIRNAME/../lib/FC_Solve/__init__.py"
import unittest
from FC_Solve import FC_Solve_Suite


class MyTests(unittest.TestCase):
    def test_null_plan(self):
        fcs = FC_Solve_Suite(self)

        name = "null plan"

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(name, None)

        # TEST
        fcs.flare_plan_num_items_is(name, 2)

        # TEST*$flare_plan_item_is_run_indef
        fcs.flare_plan_item_is_run_indef(name, 0, 0)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(name + " No. 1", 1)

    def test_empty_plan(self):
        fcs = FC_Solve_Suite(self)

        name = "empty string plan"

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(name, "")

        # TEST
        fcs.flare_plan_num_items_is(name, 2)

        # TEST*$flare_plan_item_is_run_indef
        fcs.flare_plan_item_is_run_indef(name, 0, 0)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(name + " No. 1", 1)

    def test_two_runs(self):
        fcs = FC_Solve_Suite(self)

        name = "Two Run's"
        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "Input Flares",
            ["--flare-name", "dfs", "-nf",
             "--flare-name", "befs", "--method", "a-star"])

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(name, "Run:500@befs,Run:1500@dfs")

        # 2 runs and then the implicit checkpoint.
        # TEST
        fcs.flare_plan_num_items_is(name, 3)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(name + " No. 0", 0, 1, 500)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(name + " No. 1", 1, 0, 1500)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(name + " No. 2", 2)

    def test_with_checkpoints(self):
        testname = "With checkpoints"

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "Input Flares",
            ["--flare-name", "dfs", "-nf",
             "--flare-name", "befs", "--method", "a-star", "-nf",
             "--flare-name", "foo", "--method", "a-star",
                "-asw", "0.2,0.3,0.5,0,0", "-nf",
             "--flare-name", "bar", "-to", "[01][23467]",
             ])

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(
            testname,
            "Run:500@dfs,Run:300@bar,CP:,Run:1000@befs"
        )

        # 4 items and then the implicit checkpoint.
        # TEST
        fcs.flare_plan_num_items_is(testname, 5)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 0" % (testname)), 0, 0, 500)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 1" % (testname)), 1, 3, 300)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 2" % (testname)), 2)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 3" % (testname)), 3, 1, 1000)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 4" % (testname)), 4)

    def test_with_checkpoints_and_explicit_checkpoint(self):
        testname = "With checkpoints with explicit checkpoint at end."

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "Input Flares",
            ["--flare-name", "dfs", "-nf",
             "--flare-name", "befs", "--method", "a-star", "-nf",
             "--flare-name", "foo", "--method", "a-star",
                "-asw", "0.2,0.3,0.5,0,0", "-nf",
             "--flare-name", "bar", "-to", "[01][23467]",
             ])

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(
            testname,
            "Run:500@dfs,Run:300@bar,CP:,Run:1000@befs,CP:"
        )

        # 5 items (without an extra and redundant explicit
        # checkpoint at the end.)
        # TEST
        fcs.flare_plan_num_items_is(testname, 5)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 0" % (testname)), 0, 0, 500)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 1" % (testname)), 1, 3, 300)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 2" % (testname)), 2)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 3" % (testname)), 3, 1, 1000)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 4" % (testname)), 4)

    def test_with_run_indef(self):
        testname = "With RunIndef"

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "Input Flares",
            ["--flare-name", "dfs", "-nf",
             "--flare-name", "befs", "--method", "a-star", "-nf",
             "--flare-name", "foo", "--method", "a-star",
                "-asw", "0.2,0.3,0.5,0,0", "-nf",
             "--flare-name", "bar", "-to", "[01][23467]",
             ])

        # TEST*$compile_flares_plan_ok
        fcs.compile_flares_plan_ok(
            testname,
            "Run:500@dfs,Run:300@bar,CP:,RunIndef:befs"
        )

        # 5 items (with an extra implicit chechpoint at the end).
        # TEST
        fcs.flare_plan_num_items_is(testname, 5)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 0" % (testname)), 0, 0, 500)

        # TEST*$flare_plan_item_is_run
        fcs.flare_plan_item_is_run(("%s No. 1" % (testname)), 1, 3, 300)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 2" % (testname)), 2)

        # TEST*$flare_plan_item_is_run_indef
        fcs.flare_plan_item_is_run_indef(("%s No. 3" % (testname)), 3, 1)

        # TEST*$flare_plan_item_is_checkpoint
        fcs.flare_plan_item_is_checkpoint(("%s No. 4" % (testname)), 4)


if __name__ == "__main__":
    # plan(73)
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)



#!/usr/bin/env python3

from cffi import FFI
import platform

ffi = FFI()
lib = ffi.dlopen("./libfcs_find_deal." +
                 ("dll" if (platform.system() == 'Windows') else "so"))

ffi.cdef('''
long long fc_solve_find_deal_in_range(
const long long start, const long long end, const uint_fast32_t *const ints);
void * fc_solve_user__find_deal__alloc();
void fc_solve_user__find_deal__free(void *);
void fc_solve_user__find_deal__fill(void *, const char *);
const char * fc_solve_user__find_deal__run(void *, const char *, const char *);
''')



#!/usr/bin/env python3
#
# make_pysol_freecell_board.py - generate PySol's solitaire deals.
#
# Copyright by Shlomi Fish, 2018
#
# Licensed under the MIT/Expat License.
#
# This code is inspired by PySol by Markus F.X.J. Oberhumer and others.
# See: http://pysolfc.sourceforge.net/ .
import re


class Card:
    ACE = 1
    KING = 13

    def __init__(self, id, rank, suit):
        self.id, self.rank, self.suit = id, rank, suit
        self.empty, self.flipped = False, False

    def suit_s(self):
        return 'CSHD'[self.suit]

    def is_ace(self):
        return self.rank == self.ACE

    def is_king(self):
        return self.rank == self.KING

    def flip(self, flipped=True):
        ret = Card(self.id, self.rank, self.suit)
        ret.flipped = flipped
        return ret


class CardRenderer:
    """docstring for CardRenderer"""
    def __init__(self, print_ts):
        self.print_ts = print_ts

    def to_s(self, card):
        if card.empty:
            return '-'
        ret = self.rank_s(card) + card.suit_s()
        if card.flipped:
            ret = '<' + ret + '>'
        return ret

    def found_s(self, card):
        return card.suit_s() + '-' + self.rank_s(card)

    def rank_s(self, card):
        ret = "0A23456789TJQK"[card.rank]
        if ((not self.print_ts) and ret == 'T'):
            ret = '10'
        return ret

    def render_l(self, lst):
        return [self.to_s(x) for x in lst]

    def l_concat(self, lst):
        return ' '.join(self.render_l(lst))


def createCards(num_decks, max_rank=13):
    ret = []
    for _ in range(num_decks):
        id = 0
        for s in range(4):
            for r in range(max_rank):
                ret.append(Card(id, r+1, s))
                id += 1
    return ret


def ms_rearrange(cards):
    if len(cards) != 52:
        return cards
    c = []
    for i in range(13):
        for j in (0, 39, 26, 13):
            c.append(cards[i + j])
    return c


def find_index__board_string_to_ints(content):
    if len(content) > 0 and content[-1] != '\n':
        content += '\n'

    rank_s = 'A23456789TJQK'
    rank_re = r'[' + rank_s + r']'
    suit_s = 'CSHD'
    suit_re = r'[' + suit_s + r']'

    card_re = rank_re + suit_re
    card_re_paren = r'(' + card_re + r')'

    def string_repeat_for_transcrypt(s, n):
        if n == 0:
            return ''
        return s + string_repeat_for_transcrypt(s, n-1)

    def make_line(n):
        return r':?[ \t]*' + card_re_paren + \
            string_repeat_for_transcrypt((r'[ \t]+' + card_re_paren), n-1) + \
            r'[ \t]*\n'

    complete_re = r'^' + string_repeat_for_transcrypt(make_line(7), 4) + \
        string_repeat_for_transcrypt(make_line(6), 4) + '\s*$'

    m = re.match(complete_re, content)
    if not m:
        raise ValueError("Could not match.")

    cards = CardRenderer(True).render_l(ms_rearrange(createCards(1)))

    # Reverse shuffle:
    ints = []
    limit = n = 4 * 13 - 1
    for i in range(limit):
        col = i // 8
        row = i % 8
        s = m.group(1 + col + (4*7+(row-4)*6 if row >= 4 else row*7))
        idx = [j for j in range(n+1) if cards[j] == s]
        if len(idx) != 1:
            raise ValueError("Foo")
        j = idx[0]
        ints.append(j)
        cards[j] = cards[n]
        n -= 1

    return ints


def factory():
    return {'find_index__board_string_to_ints':
            find_index__board_string_to_ints}



#!/usr/bin/env python3
#
# make_pysol_freecell_board.py - generate PySol's solitaire deals.
#
# Copyright by Shlomi Fish, 2018
#
# Licensed under the MIT/Expat License.
#
# This code is inspired by PySol by Markus F.X.J. Oberhumer and others.
# See: http://pysolfc.sourceforge.net/ .
import random2
import sys
from fc_solve_find_index_s2ints import createCards, Card, CardRenderer, \
        find_index__board_string_to_ints, ms_rearrange


def empty_card():
    ret = Card(0, 0, 0)
    ret.empty = True
    return ret


class RandomBase:
    DEALS_PYSOL = 0
    DEALS_PYSOLFC = 1
    DEALS_MS = 2

    def shuffle(self, seq):
        for n in range(len(seq)-1, 0, -1):
            j = self.randint(0, n)
            seq[n], seq[j] = seq[j], seq[n]
        return seq

    def randint(self, a, b):
        return a + int(self.random() * (b+1-a))


class PysolRandom(RandomBase):
    def setSeed(self, seed):
        seed = self._convertSeed(seed)
        if not 0 <= seed <= self.MAX_SEED:
            raise ValueError("seed is out of range")
        self.seed = seed
        return seed

    def _convertSeed(self, seed):
        return int(seed)

# /***********************************************************************
# // Linear Congruential random generator
# //
# // Knuth, Donald.E., "The Art of Computer Programming,", Vol 2,
# // Seminumerical Algorithms, Third Edition, Addison-Wesley, 1998,
# // p. 106 (line 26) & p. 108
# ************************************************************************/


class LCRandom64(PysolRandom):
    MAX_SEED = 0xffffffffffffffff  # 64 bits

    def random(self):
        self.seed = (self.seed*6364136223846793005 + 1) & self.MAX_SEED
        return ((self.seed >> 21) & 0x7fffffff) / 2147483648.0


class LCRandom31(RandomBase):
    MAX_SEED = ((1 << (32+2))-1)         # 34 bits

    def setSeed(self, seed):
        self.seed = seed
        self.seedx = seed if (seed < 0x100000000) else (seed - 0x100000000)

    def random(self):
        if (self.seed < 0x100000000):
            ret = self._rand()
            return (ret if (self.seed < 0x80000000) else (ret | 0x8000))
        else:
            return self._randp() + 1

    def _randp(self):
        self.seedx = ((self.seedx) * 214013 + 2531011) & self.MAX_SEED
        return (self.seedx >> 16) & 0xffff

    def _rand(self):
        self.seedx = ((self.seedx) * 214013 + 2531011) & self.MAX_SEED
        return (self.seedx >> 16) & 0x7fff

    def randint(self, a, b):
        return a + self.random() % (b+1-a)


# * Mersenne Twister random number generator
class MTRandom(RandomBase, random2.Random):
    MAX_SEED = 100000000000000000000  # 20 digits

    def setSeed(self, seed):
        random2.Random.__init__(self, seed)
        self.initial_state = self.getstate()


class Columns:
    def __init__(self, num):
        self.num = num
        self.cols = [[] for _ in range(num)]

    def add(self, idx, card):
        self.cols[idx].append(card)

    def rev(self):
        self.cols.reverse()


class Board:
    def __init__(self, num_columns, with_freecells=False,
                 with_talon=False, with_foundations=False):
        self.with_freecells = with_freecells
        self.with_talon = with_talon
        self.with_foundations = with_foundations
        self.raw_foundations_line = None
        self.columns = Columns(num_columns)
        if self.with_freecells:
            self.freecells = []
        if self.with_talon:
            self.talon = []
        if self.with_foundations:
            self.foundations = [empty_card() for s in range(4)]
        self._lines = []

    def add_line(self, string):
        self._lines.append(string)

    def reverse_cols(self):
        self.columns.rev()

    def add(self, idx, card):
        self.columns.add(idx, card)

    def add_freecell(self, card):
        if not self.with_freecells:
            raise AttributeError("Layout does not have freecells!")
        self.freecells.append(card)

    def add_talon(self, card):
        if not self.with_talon:
            raise AttributeError("Layout does not have a talon!")
        self.talon.append(card)

    def put_into_founds(self, card):
        if not self.with_foundations:
            raise AttributeError("Layout does not have foundations!")
        res = self.foundations[card.suit].rank+1 == card.rank
        if res:
            self.foundations[card.suit] = card
        return res

    def print_foundations(self, renderer):
        cells = []
        for f in [2, 0, 3, 1]:
            if not self.foundations[f].empty:
                cells.append(renderer.found_s(self.foundations[f]))

        if len(cells):
            self.add_line("Foundations:" + ("".join([" "+s for s in cells])))

    def gen_lines(self, renderer):
        self._lines = []
        if self.with_talon:
            self.add_line("Talon: " + renderer.l_concat(self.talon))
        if self.with_foundations:
            self.print_foundations(renderer)
        if self.raw_foundations_line:
            self.add_line(self.raw_foundations_line)
        if self.with_freecells:
            self.add_line("Freecells: " + renderer.l_concat(self.freecells))

        self._lines += [renderer.l_concat(c) for c in self.columns.cols]

    def calc_string(self, renderer):
        self.gen_lines(renderer)
        return "".join(l + "\n" for l in self._lines)


def shuffle(cards, game_num, which_deals):
    ms = ((game_num <= 32000) or (which_deals == RandomBase.DEALS_MS))
    r = LCRandom31() if ms else MTRandom() if which_deals == \
        RandomBase.DEALS_PYSOLFC else LCRandom64()
    r.setSeed(game_num)
    return r.shuffle(ms_rearrange(cards) if ms else cards)


class Game:
    REVERSE_MAP = \
        {
                "freecell":
                ["forecell", "bakers_game",
                 "ko_bakers_game", "kings_only_bakers_game",
                 "relaxed_freecell", "eight_off"],
                "der_katz":
                ["der_katzenschwantz", "die_schlange"],
                "seahaven":
                ["seahaven_towers", "relaxed_seahaven",
                 "relaxed_seahaven_towers"],
                "bakers_dozen": None,
                "gypsy": None,
                "klondike":
                ["klondike_by_threes",
                 "casino_klondike", "small_harp", "thumb_and_pouch",
                 "vegas_klondike", "whitehead"],
                "simple_simon": None,
                "yukon": None,
                "beleaguered_castle":
                ["streets_and_alleys", "citadel"],
                "fan": None,
                "black_hole": None,
                "all_in_a_row": None,
        }

    def __init__(self, game_id, game_num, which_deals, max_rank=13):
        mymap = {}
        for k, v in self.REVERSE_MAP.items():
            for name in [k] + (v if v else []):
                mymap[name] = k
        self.games_map = mymap
        self.game_id = game_id
        self.game_num = game_num
        self.which_deals = which_deals
        self.max_rank = max_rank
        self.game_class = self.games_map[self.game_id]
        if not self.game_class:
            raise ValueError("Unknown game type " + self.game_id + "\n")

    def is_two_decks(self):
        return self.game_id in ("der_katz", "der_katzenschwantz",
                                "die_schlange", "gypsy")

    def get_num_decks(self):
        return 2 if self.is_two_decks() else 1

    def calc_deal_string(self, game_num, renderer):
        self.game_num = game_num
        self.deal()
        getattr(self, self.game_class)()
        return self.board.calc_string(renderer)

    def calc_layout_string(self, renderer):
        self.deal()
        getattr(self, self.game_class)()
        return self.board.calc_string(renderer)

    def print_layout(self, renderer):
        print(self.calc_layout_string(renderer), sep='', end='')

    def new_cards(self, cards):
        self.cards = cards
        self.card_idx = 0

    def deal(self):
        cards = shuffle(createCards(self.get_num_decks(),
                                    self.max_rank),
                        self.game_num, self.which_deals)
        cards.reverse()
        self.new_cards(cards)

    def __iter__(self):
        return self

    def no_more_cards(self):
        return self.card_idx >= len(self.cards)

    def __next__(self):
        if self.no_more_cards():
            raise StopIteration
        c = self.cards[self.card_idx]
        self.card_idx += 1
        return c

    def add(self, idx, card):
        self.board.add(idx, card)

    def add_freecell(self, card):
        self.board.add_freecell(card)

    def cyclical_deal(self, num_cards, num_cols, flipped=False):
        for i in range(num_cards):
            self.add(i % num_cols, next(self).flip(flipped=flipped))

    def add_all_to_talon(self):
        for c in self:
            self.board.add_talon(c)

    def add_empty_fc(self):
        self.add_freecell(empty_card())

    def _shuffleHookMoveSorter(self, cards, cb, ncards):
        extracted, i, new = [], len(cards), []
        for c in cards:
            select, ord_ = cb(c)
            if select:
                extracted.append((ord_, i, c))
                if len(extracted) >= ncards:
                    new += cards[(len(cards)-i+1):]
                    break
            else:
                new.append(c)
            i -= 1
        return new, [x[2] for x in reversed(sorted(extracted))]

    def _shuffleHookMoveToBottom(self, inp, cb, ncards=999999):
        cards, scards = self._shuffleHookMoveSorter(inp, cb, ncards)
        return scards + cards

    def _shuffleHookMoveToTop(self, inp, cb, ncards=999999):
        cards, scards = self._shuffleHookMoveSorter(inp, cb, ncards)
        return cards + scards

    def all_in_a_row(game):
        game.board = Board(13)
        game.cards = game._shuffleHookMoveToTop(
            game.cards,
            lambda c: (c.id == 13, c.suit),
            1)
        game.cyclical_deal(52, 13)
        game.board.raw_foundations_line = 'Foundations: -'

    def bakers_dozen(game):
        n = 13
        cards = list(reversed(game.cards))
        for i in [i for i, c in enumerate(cards) if c.is_king()]:
            j = i % n
            while j < i:
                if not cards[j].is_king():
                    cards[i], cards[j] = cards[j], cards[i]
                    break
                j += n
        game.new_cards(cards)
        game.board = Board(13)
        game.cyclical_deal(52, 13)

    def beleaguered_castle(game):
        game.board = Board(8, with_foundations=True)
        if game.game_id in ('beleaguered_castle', 'citadel'):
            new = []
            for c in game:
                if c.is_ace():
                    game.board.put_into_founds(c)
                else:
                    new.append(c)
            game.new_cards(new)
        for _ in range(6):
            for s in range(8):
                c = next(game)
                if not ((game.game_id == 'citadel') and
                        game.board.put_into_founds(c)):
                    game.add(s, c)
            if game.no_more_cards():
                break
        if game.game_id == 'streets_and_alleys':
            game.cyclical_deal(4, 4)

    def black_hole(game):
        game.board = Board(17)
        game.cards = game._shuffleHookMoveToBottom(
            game.cards,
            lambda c: (c.id == 13, c.suit),
            1)
        next(game)
        game.cyclical_deal(52-1, 17)
        game.board.raw_foundations_line = 'Foundations: AS'

    def der_katz(game):
        is_ds = game.game_id == 'die_schlange'
        if is_ds:
            print('Foundations: H-A S-A D-A C-A H-A S-A D-A C-A')
        game.board = Board(9)
        i = 0
        for c in game:
            if c.is_king():
                i += 1
            if not (is_ds and c.is_ace()):
                game.add(i, c)

    def fan(game):
        game.board = Board(18)
        game.cyclical_deal(52-1, 17)
        game.add(17, next(game))

    def freecell(game):
        is_fc = (game.game_id in ("forecell", "eight_off"))
        game.board = Board(8, with_freecells=is_fc)
        game.cyclical_deal(4 *
                           (game.max_rank-1 if is_fc else game.max_rank), 8)

        if is_fc:
            for c in game:
                game.add_freecell(c)
                if game.game_id == "eight_off":
                    game.add_empty_fc()

    def gypsy(game):
        num_cols = 8
        game.board = Board(num_cols, with_talon=True)
        game.cyclical_deal(num_cols*2, num_cols, flipped=True)
        game.cyclical_deal(num_cols, num_cols)
        game.add_all_to_talon()

    def klondike(game):
        num_cols = 7
        game.board = Board(num_cols, with_talon=True)
        for r in range(num_cols-1, 0, -1):
            game.cyclical_deal(r, r, flipped=True)
        game.cyclical_deal(num_cols, num_cols)
        game.add_all_to_talon()
        if not (game.game_id == 'small_harp'):
            game.board.reverse_cols()

    def seahaven(game):
        game.board = Board(10, with_freecells=True)
        game.cyclical_deal(50, 10)
        game.add_empty_fc()
        for c in game:
            game.add_freecell(c)

    def simple_simon(game):
        game.board = Board(10)
        for num_cards in range(9, 2, -1):
            game.cyclical_deal(num_cards, num_cards)
        game.cyclical_deal(10, 10)

    def yukon(game):
        num_cols = 7
        game.board = Board(num_cols)
        for i in range(1, num_cols):
            for j in range(i, num_cols):
                game.add(j, next(game).flip())
        for i in range(4):
            for j in range(1, num_cols):
                game.add(j, next(game))
        game.cyclical_deal(num_cols, num_cols)


def make_pysol_board__main(args):
    print_ts = False
    which_deals = RandomBase.DEALS_PYSOL
    max_rank = 13
    while args[1][0] == '-':
        a = args[1]
        if a == "-t":
            print_ts = True
            args.pop(1)
        elif (a == "--max-rank"):
            args.pop(1)
            max_rank = int(args.pop(1))
        elif (a == "--pysolfc") or (a == "-F"):
            which_deals = RandomBase.DEALS_PYSOLFC
            args.pop(1)
        elif (a == "--ms") or (a == "-M"):
            which_deals = RandomBase.DEALS_MS
            args.pop(1)
        else:
            raise ValueError("Unknown flag " + a + "!")

    game_num = int(args[1])
    which_game = args[2] if len(args) >= 3 else "freecell"
    Game(which_game, game_num, which_deals, max_rank).print_layout(
        CardRenderer(print_ts))


def find_index_main(args, find_ret):
    output_to_stdout = True
    output_fn = None
    is_ms = False
    while len(args) >= 2 and args[1][0] == '-':
        if args[1] == "-o":
            args.pop(1)
            if not len(args):
                raise ValueError("-o must accept an argument.")
            output_to_stdout = False
            output_fn = args.pop(1)
        elif args[1] == '--ms':
            args.pop(1)
            is_ms = True
        elif args[1] == '-':
            break
        else:
            raise ValueError("Unknown flag " + args[1] + "!")

    if not is_ms:
        raise ValueError("only --ms is supported for now!")
    input_from_stdin = True
    input_fn = None
    if len(args) >= 2:
        if args[1] != "-":
            input_fn = args.pop(1)
            input_from_stdin = False

    content = []
    if input_from_stdin:
        content = sys.stdin.readlines()
    else:
        with open(input_fn) as f:
            content = f.readlines()

    content = ''.join(content)
    ints = find_index__board_string_to_ints(content)
    ret = find_ret(ints)

    ret_code = 0
    if ret >= 0:
        s = "Found deal = %d" % ret
        if output_to_stdout:
            print(s)
        else:
            with open(output_fn, 'w') as f:
                f.write("%s\n" % s)
        ret_code = 0
    else:
        print("Not found!")
        ret_code = -1
    return ret_code


if __name__ == "__main__":
    sys.exit(make_pysol_board__main(sys.argv))



#!/usr/bin/env python3

from FC_Solve import FC_Solve_Suite
import unittest
# TEST:source "$^CURRENT_DIRNAME/../lib/FC_Solve/__init__.py"


class MyTests(unittest.TestCase):
    def test_main(self):
        fcs = FC_Solve_Suite(self)
        # TEST*$test_befs
        fcs.test_befs_weights(
            "Simple - all integers",
            "5,4,3,0,2",
            [5.0, 4.0, 3.0, 0.0, 2.0])

        # TEST*$test_befs
        fcs.test_befs_weights(
            "Simple - fractions",
            "0.2,0.3,0.4,0.5,0.6",
            [0.2, 0.3, 0.4, 0.5, 0.6])

        # TEST*$test_befs
        fcs.test_befs_weights(
            "Testing end after three numbers",
            "0.2,0.35,0.4\x00500\x00600.7\x00",
            [0.2, 0.35, 0.4, 0, 0])

        # TEST*$test_befs
        fcs.test_befs_weights(
            "One number with a trailing comma",
            "100,\x002,3,4,5,6,7,8,9,100\x00",
            [100, 0, 0, 0, 0])


if __name__ == "__main__":
    # plan(24)
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)



#!/usr/bin/env python3

import subprocess
import os
import re
import unittest

prog = os.environ['FCS_PATH'] + '/fc-solve'


def _normalize_lf(txt):
    return txt.replace("\r\n", "\n")


class MyTests(unittest.TestCase):
    def mytest(self, cmd, input_text, want_prefix, msg):
        process = subprocess.Popen(
            cmd,
            shell=False, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        process.stdin.write(bytes(input_text, 'utf_8'))
        process.stdin.close()
        got_out = _normalize_lf(process.stdout.read().decode('utf-8'))
        self.assertTrue(re.match(want_prefix, got_out), msg)

    def test_main(self):
        # plan(6)

        # TEST
        self.mytest(
            [prog, '-i', '--iter-output-step', '100', '-mi', '1000', '-'],
            """9D 6S 4D AH 5C 4H KH
JH JC JS 7S 6D AC 4S
AD 8D AS 2H KC KD 7H
KS TS 9S 8H JD QD 3H
TC 4C 2S 5D 3S 6H
3C QC 5S QS QH 7D
9H 5H 7C 6C 3D TD
2C 8C 8S TH 2D 9C
""",
            ('^' + "".join([(("Iteration: %d\nDepth: [0-9]+\n" +
                            "Stored-States: [0-9]+\nScan:[^\\n]+\n\n") %
                             (i)) for i in [0, 100, 200, 300]])),
            "Freecell MS deal 1941 initial",
        )


if __name__ == '__main__':
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)



#!/usr/bin/env python3
#
# find-freecell-deal-index.py - a program to find out the Freecell deal index
# based on the initial cards layout.
#
# Copyright by Shlomi Fish, 2018
#
# Licensed under the MIT/Expat License.

import sys
from make_pysol_freecell_board import find_index_main, LCRandom31
from six.moves import range


def find_ret(ints):
    def is_right(d):
        r = LCRandom31()
        r.setSeed(d)
        n = len(ints)
        for i in ints:
            if r.randint(0, n) != i:
                return False
            n -= 1
        return True

    for d in range(1, (1 << 33)):
        if is_right(d):
            return d
    return -1


if __name__ == "__main__":
    sys.exit(find_index_main(sys.argv, find_ret))



#!/usr/bin/env python3

from FC_Solve import FC_Solve_Suite
import unittest
# TEST:source "$^CURRENT_DIRNAME/../lib/FC_Solve/__init__.py"


class MyTests(unittest.TestCase):
    def test_1(self):
        testname = "test_1"

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "dto 1",
            ["-to", "0123456789", "-dto", "1,0132456789"]
        )

        # TEST
        fcs.num_by_depth_tests_order_is(testname, 2)

        # TEST
        fcs.by_depth_max_depth_of_depth_idx_is(testname, 0, 1)

    def test_2(self):
        testname = "test_1"

        fcs = FC_Solve_Suite(self)

        # TEST*$input_cmd_line
        fcs.input_cmd_line(
            "dto 1",
            ["-to", "0123456789", "-dto", "1,0132456789", "-dto", "50,056234"]
        )

        # TEST
        fcs.num_by_depth_tests_order_is(testname, 3)

        # TEST
        fcs.by_depth_max_depth_of_depth_idx_is(testname, 0, 1)

        # TEST
        fcs.by_depth_max_depth_of_depth_idx_is(testname, 1, 50)


if __name__ == "__main__":
    # plan(7)
    from pycotap import TAPTestRunner
    suite = unittest.TestLoader().loadTestsFromTestCase(MyTests)
    TAPTestRunner().run(suite)
